# BGP.pm - module implementating the BGP protocol (IETF RFC1771)
# Copyright (C) 2001 Steven Hessing 
# steven@xs4all.nl

# This software is distributed under the terms of the license described in the
# file `LICENSE'. If this file is not included in the distribution then
# please contact the author of this software.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

package BGP;

require Exporter;
@ISA = (Exporter);

use strict;
use warnings;

use POSIX qw(strftime);
use Socket;
use Sys::Hostname;
use Sys::Syslog qw(:DEFAULT setlogsock);
use FileHandle;

use BGP::Constants;
use BGP::Neighbor;
use BGP::PathAttribute;
use BGP::AdjRibIn;
use BGP::LocalRib;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {};
	bless $self, $class;

	my ($configfile, $logtype) = @_;

	$self->set_beacon_prefix("10.0.0.0/8"); #default beacon prefix

	$self->set_seqNo(0);
	$self->set_ASseqNo(64511);

	if (!$self->read_config_file($configfile)) {
		die "Unable to read $configfile: $!\n";
	}

	#command line overrides config file "log file <filename>"
	$self->set_logging ($logtype) unless ($logtype eq ""); 
	if(!defined($$self{loghandle})) {$self->set_logging("");}

	my $hostname = hostname();
	my ($name, $aliases, $type, $len, $clientaddr) = gethostbyname($hostname);

	die "unable to get my ip address in init_bgp_socket()" if ($name eq '');

	$$self{filehandle} = new FileHandle;

	my $proto = getprotobyname ('tcp');
	socket($$self{filehandle},AF_INET,SOCK_STREAM,$proto) || die "socket: $!";

	my $port = getservbyname ('bgp','tcp');
	my $sin = sockaddr_in($port, INADDR_ANY);
	setsockopt($$self{filehandle}, SOL_SOCKET, SO_REUSEADDR, pack ('l', 1));

	bind($$self{filehandle},$sin)                    || die "bind: $!";
	listen($$self{filehandle},5)                     || die "listen: $!";

	if (!defined $self->get_routerid ()) {
		$self->set_routerid ($clientaddr);
	}

	$$self{adjRIBin} = new BGP::AdjRibIn;
	$$self{localRIB} = new BGP::LocalRib;

	return $self;
}

sub DESTROY {
	my $bgprouter = shift;

	my $key;
	foreach $key ($bgprouter->get_established_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);
		$neighbor->close ($bgprouter->get_adjRIBin(), BGP_ERRC_Cease);
	}
}

sub incoming_connection {
	my $bgprouter = shift;

	my $from;
	my $tmpfh = new FileHandle;
	if (!($from=accept($tmpfh,$bgprouter->get_filehandle()))) {
		$bgprouter->log (128, 1, "incoming_connection: Can't accept: $!");
		return 0;
	}
	my ($port,$host) = sockaddr_in($from);
	my $host_ip = inet_ntoa ($host);

	if (!$bgprouter->neighbor_exists ($host_ip)) {
		$bgprouter->log (128, 16, "Incoming BGP session from $host_ip rejected");
		return $host_ip;
	}
	my $mysockaddr = getsockname ($tmpfh);
	my ($myport, $myhost) = sockaddr_in ($mysockaddr);

	$bgprouter->log (128, 4, "Established BGP session with $host_ip:$port");
	my $neighbor = $bgprouter->get_neighbor ($host_ip);

	# Nexthop here is the IP address for which we received the connection
	# request. We bind to * so we don't know what IP address the neighbor
	# used. We use this IP address as nexthop address when we send eBGP
	$neighbor->set_nexthop ($myhost);

	# Init the rest of the session data
	$neighbor->set_ipaddress ($host_ip);
	$neighbor->set_state (BGP_NEIGHBOR_ACTIVE);
	$neighbor->set_filehandle ($tmpfh);
	$neighbor->clear_routesin ();
	$neighbor->clear_routesout ();
	$neighbor->clear_routescurrent ();
	$neighbor->clear_history ();
	$neighbor->set_time ();
	$neighbor->add_history ("BGP Transport connection open");
	

	return 0;
}

sub get_adjRIBin {
	my $bgprouter = shift;

	return $$bgprouter{adjRIBin};
}

sub get_localRIB {
	my $bgprouter = shift;

	return $$bgprouter{localRIB};
}

sub get_filehandle {
	my $bgprouter = shift;

	return $$bgprouter{filehandle};
}

sub set_filehandle {
	my $bgprouter = shift;
	my $fh = shift;

	return $$bgprouter{filehandle} = $fh;
}

sub neighbor_exists {
	my $bgprouter = shift;
	my $ip = shift;

	return $bgprouter->get_neighbor ($ip);
}

sub connection_exists {
	my $bgprouter = shift;
	my $ip = shift;

	my $neighbor = $bgprouter->get_neighbor ($ip);
	return defined $neighbor->get_filehandle();
}

sub get_neighbor {
	my $self = shift;
	my $ip = shift;

	return $$self{neighbor}{$ip};
}

sub add_neighbor {
	my $self = shift;
	my $ip = shift;

	return if (defined $$self{neighbor}{$ip});

	return $$self{neighbor}{$ip} = new BGP::Neighbor ($ip);
}

sub get_neighbors {
	my $bgprouter = shift;

	return (keys %{$$bgprouter{neighbor}});
}

sub get_established_neighbors {
	my $bgprouter = shift;

	my @neighbors;
	my $key;
	foreach $key (keys %{$$bgprouter{neighbor}}) {
		my $neighbor = $bgprouter->get_neighbor ($key);
		my $state = $neighbor->get_state ();
		if (defined $state && $state == BGP_NEIGHBOR_ESTABLISHED) {
			push @neighbors, $key;
		}
	}
	return (@neighbors);
}

sub get_connected_neighbors {
	my $bgprouter = shift;

	my @neighbors;
	my $key;
	foreach $key (keys %{$$bgprouter{neighbor}}) {
		my $neighbor = $bgprouter->get_neighbor ($key);
		if (defined $neighbor->get_filehandle ()) {
			push @neighbors, $key;
		}
	}
	return (@neighbors);
}

sub get_routerid {
	my $bgprouter = shift;

	return $$bgprouter{ip};
}

sub set_routerid {
	my $bgprouter = shift;
	my $ip = shift;

	return $$bgprouter{ip} = $ip;
}

sub get_as {
	my $bgprouter = shift;

	return $$bgprouter{as};
}

sub set_as {
	my $bgprouter = shift;
	my $as = shift;

	return $$bgprouter{as} = $as;
}

sub read_config_file {
	my $bgprouter = shift;
	my ($file) = @_;

	my %bgp_log_types = (
		"OPEN" => 1,
		"UPDATE" => 2,
		"NOTIFICATION" => 4,
		"KEEPALIVE" => 8,
		"ROUTE-REFRESH" => 16,
		"LIST" => 32,
		"HEADER" => 64,
		"GENERAL" => 128,
		"LOCALRIB" => 256,
		"BEACON" => 300
	);

	my %bgp_log_levels = (
		"CRITICAL" => 1,
		"WARNING" => 2,
		"INFO" => 4,
		"PARSE" => 8,
		"SESSION-ERRORS" => 16,
		"SESSION-WARNINGS" => 32,
		"PARSE-OPEN" => 64,
		"PARSE-OPEN-OPTIONAL" => 128,
		"PARSE-UPDATE" => 64,
		"PARSE-UPDATE-PREFIX" => 128,
		"PARSE-UPDATE-PATH-ATTRIBUTES" => 256,
		"PARSE-NOTIFICATION" => 64,
		"PARSE-REFRESH" => 64,
		"SESSION-PARSE" => 64,
		"LOCALRIB-CHANGES" => 64,
		"LOCALRIB-CHANGES-ROUTES" => 128,
		"LOCALRIB-CHANGES-EVAL" => 256
	);

	my $fh = new FileHandle;
	if (!open ($fh, $file)) {
		die "Error opening file $file: $!";
	}
	my $line;
	while ($line = get_config_line ($fh)) {
$bgprouter->log(128,4,"LINE:$line");
		$line = lc ($line);
		my @words = split /\s+/, $line;

		$words[0]=~tr/A-Z/a-z/; #ignore case
$bgprouter->log(128,4,"WORD 0 is $words[0].\n");
		#beacon prefix
		if ($words[0] eq "network") {
		    $bgprouter->set_beacon_prefix($words[1]);
		}
		#local pref
		if ($words[0] eq "local-pref") {
		    $bgprouter->set_local_pref($words[1]);
		}
		#MED
		if ($words[0] eq "med") {
		    $bgprouter->set_MED($words[1]);
		}
		#community
		if ($words[0] eq "community") {
		    $bgprouter->set_community(@words);
		}


		if ($words[0] eq "log" && $words[1] eq "file") {
		    $bgprouter->set_logging($words[2]);
		}

		elsif ($words[0] eq "log") {
			my $type = uc ($words[1]);
			my $level = uc ($words[2]);
			if (defined $bgp_log_types{$type} &&
					defined $bgp_log_levels{$level} ) {
				$bgprouter->set_loglevel ($bgp_log_types{$type},
					$bgp_log_levels{$level});
			}
		}
		if ($words[0] eq "router" && $words[1] eq "bgp") {
			$bgprouter->set_as ($words[2]);
		}
		if ($words[0] eq "bgp" && $words[1] eq "confederation") {
		    if($words[2] eq "identifier") {
			$bgprouter->set_confedID($words[3]);
		    }elsif($words[2] eq "peers") {
			$bgprouter->set_confedPeers(@words);
		    }
		}

		if ($words[0] eq "router-id") {
		        $bgprouter->set_routerid (inet_aton($words[1]));
		}
		if ($words[0] eq "neighbor") {
			my $ip = $words[1];
			my $neighbor = $bgprouter->get_neighbor ($ip);

			$bgprouter->log(128,4,"configfile neighbor: $ip.\n");

			if (!defined $neighbor) {
				$neighbor = $bgprouter->add_neighbor ($ip);
				$neighbor->set_ipaddress ($ip);
				$neighbor->set_log_filehandle
					($bgprouter->get_log_filehandle ());
				foreach my $type ($bgprouter->get_defined_loglevels ()) {
					my $level = $bgprouter->get_loglevel ($type);
					if (defined $level) {
						$neighbor ->set_loglevel ($type, $level);
					}
				}
			}
			if ($words[2] eq "remote-as") {
				my $as = $words[3];
				$neighbor->set_as ($as);
			}
		}
	}
	close $fh;
	return 1;
}

sub get_config_line {
    my ($fh) = @_;

    my $line = "";

    while ($line = <$fh>) {
	chomp($line);
	# Remove whitespace at the beginning
	if ($line =~ /^\s+(.*)$/) {
	    $line = $1;
	}
	# if line is empty then skip it
	if ($line =~ /^\s*$/) {
	    next;
	}
	# Remove comments
	if($line=~/^(#)/) {next;}
#		if ($line =~ /^\s*(.*?)#.*$/) {
	if ($line =~ /^(.+)#.*$/) {
	       $line = $1;
	}

	   # Read next line if this one ends with the continuation char: \
#		if ($line =~ /^\s*(\S.*)\\\s*$/) {
#			$line = "$1 ";
#			next;
#		}

#	my $logm = $line; chop ($logm);
	return $line;
    }
    return undef;
}

sub process_keepalives {
	my $bgprouter = shift;

	# Send out keepalives and calculate time until we need to do it again.
	my $sleep_time = $bgprouter->process_keepalives_out ();
	# Have all our neighbors sent in their keepalives on time?
	$bgprouter->process_keepalives_in ();

	return $sleep_time;
}

sub setup_select {
	my $bgprouter = shift;

	my $rin = '';
	vec($rin,fileno($bgprouter->get_filehandle()),1) = 1;
	my $key;
	foreach $key ($bgprouter->get_connected_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);
		vec($rin, fileno($neighbor->get_filehandle()),1) = 1;
	}
	return $rin;
}

sub process_select {
	my $bgprouter = shift;
	my $rout = shift;

	my $count = 0;

	if (vec($rout,fileno($bgprouter->get_filehandle()),1)) {
		# We have a new incoming TCP connection.
		$bgprouter->incoming_connection();
		$count++;
	}
	my $key;
	foreach $key ($bgprouter->get_connected_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);
		$bgprouter->log (128, 4, "Processing input for: $key");
		if (vec ($rout, fileno ($neighbor->get_filehandle()) ,1) ) {
			# here we receive all BGP messages and maintain
			# $$bgprouter{adjRIBin}{ip}}, {removed} and {todo}
			# TODO: it is ugly that we send the bgprouter object
			my ($errorcode, $errorsubcode, $errordata) = 
				$neighbor->receive ($bgprouter);
			$count++;
			if ($errorcode) {
				$neighbor->close ($bgprouter->get_adjRIBin(), $errorcode,$errorsubcode, $errordata);
			}
		}
	}
	return $count;
}

sub maintain_RIBs {
	my $bgprouter = shift;

	my $localRIB = $bgprouter->get_localRIB ();
	$localRIB->maintain ($bgprouter);
	# adjRIBin doesn't need to be maintained here
	# adjRIBout still needs to be written
}

# dump_routingtable
# The routing table is currently implemented using hashes. This will be
# -very- slow for large routing tables. But the Tree::Radix implementation
# under http://www.linnaean.org/~hag/perl/radix/radix.html has not been
# released yet. We'll also need to look at
# http://www.nada.kth.se/~snilsson/public/code/router/ which describes an
# algoritm implemented in C or java that may be faster but which needs to
# be ported to perl5.

sub dump_routingtable {
	my $bgprouter = shift;

	my $adjRIBin = $bgprouter->get_adjRIBin();
	my $localRIB = $bgprouter->get_localRIB();

	my @origin = ('i', 'e', '?');

	my ($prefix_n, $len);
	foreach $prefix_n ($adjRIBin->get_prefixes()) {
		foreach $len (sort {$b <=> $a}
				$adjRIBin->get_prefix_lengths($prefix_n)) {
			my ($nexthop_n, $bestneighbor_n, $bestprefix_n, $bestlen) =
				$localRIB->lookup_route ($prefix_n, $len);
			if (!defined $nexthop_n) {
				$bgprouter->log (256,1,"No route in local-RIB for",
				inet_ntoa($prefix_n), "/$len");
				next;
			} 
			my $neighbor_n; 
			foreach $neighbor_n ($adjRIBin->get_neighbors ($prefix_n, $len)) {
				if ($neighbor_n eq $bestneighbor_n &&
						$prefix_n eq $bestprefix_n &&
						$len == $bestlen) {
					$bgprouter->lograw ("*");
				} else {
					$bgprouter->lograw (" ");
				}
				my $pa = $adjRIBin->get_pathattribute
					($prefix_n, $len, $neighbor_n);
				my $origin = $pa->get_origin ();
				my $nexthop = inet_ntoa ($pa->get_nexthop());
				my $local_pref = $pa->get_localpref ();
				my $med = $pa->get_med ();

				$bgprouter->lograw (sprintf 
					"%1s %15s/%-2s %15s %8s %8s", $origin[$origin],
					inet_ntoa ($prefix_n), $len, $nexthop,
					defined $local_pref ?  $local_pref : "undef",
					defined $med ?  $med : "undef");
				my @aspath = $pa->get_aspath ();
				my $aspath;
				foreach $aspath (@aspath) {
					$bgprouter->lograw (
						" (" . BGP_AS_PATH->[$pa->get_pathtype($aspath)].")");
					my @aslist = $pa->get_aslist ($aspath);
					my $as;
					foreach $as (@aslist) {
						$bgprouter->lograw (" $as");
					}
				}
				$bgprouter-> lograw (" ");
				my @comms = $pa->get_communities ();
				my $comm;
				foreach $comm (@comms) {
					my $up = $comm >> 16;
					my $low = $comm & 0xFFFF;
					$bgprouter->lograw ("$up:$low ");
				}
				$bgprouter->lograw ("\n");
			}
		}
	}
}


#initialize connection opens
# sub init_open {
#     print "Init_open called\n";
#     my $bgprouter = shift;
#     #send open messages:
#     my ($key, $socket, $hostip);
#     foreach $key ($bgprouter->get_neighbors()) {
# 	my $neighbor = $bgprouter->get_neighbor($key);
# 	my $state = $neighbor->get_state ();
# 	my $my_ip = $bgprouter->get_routerid ();
# 	my $my_as = $bgprouter->get_as ();
# 	print "Here: $state: ".BGP_NEIGHBOR_IDLE."\n";
# 	if($state == BGP_NEIGHBOR_IDLE) {
	    
# 	    $socket = IO::Socket::INET->new(PeerAddr => $neighbor->get_ipaddress(),
# 					    PeerPort => 179, Proto => "tcp",
# 					    Type => SOCK_STREAM);
#             if(!$socket){
# 		my $hostip = $neighbor->get_ipaddress();
# 		print "Couldn't connect to Neighbor $hostip:179 : $!\n"; next;
# 	    }
# 	    $neighbor->set_filehandle($socket);
# 	    $neighbor->send_open ($my_ip, $my_as);
# 	    $neighbor->set_state (BGP_NEIGHBOR_ESTABLISHED);
# 	}
#     }
# print "init_open exited\n";
# }	

sub dump_neighbors {
	my $bgprouter = shift;

	my $key;
	foreach $key ($bgprouter->get_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);

		my $peer_ip = $neighbor->get_ipaddress ();
		my $as = $neighbor->get_as ();
		my $rin = $neighbor->get_routesin ();
		my $rout = $neighbor->get_routesout ();
		my $rcur = $neighbor->get_routescurrent ();
		my $rib = $neighbor->get_localRIBentries ();
		my $state = $neighbor->get_state ();
		my $t = $neighbor->get_time ();
		if (defined $t) {
			$t = strftime ("%D %T", localtime ($t));
		} else {
			$t = "never";
		}

		$bgprouter->lograw (sprintf
			"Neighbor %s AS: %d Status: %s (since %s)\n",
			$peer_ip, $as, BGP_NEIGHBOR_STATUS -> [$state], $t);
		$bgprouter->lograw (sprintf
			"Routes: %d recv, %d sent, %d cur, %d localRIB\n",
			$rin, $rout, $rcur, $rib);

		if ($state == BGP_NEIGHBOR_ESTABLISHED) {
			$bgprouter->lograw (sprintf
				"Holdtime: %3d Keepalive: %18s in, %18s out\n",
				$neighbor->get_holdtime (),
				strftime("%D %T",
				localtime($neighbor->get_timer_keepalivein ())),
				strftime("%D %T",
				localtime($neighbor->get_timer_keepaliveout ())));
		}
		my $hist;
		foreach $hist ($neighbor->get_history()) {
		my ($ti, $event) = $neighbor->get_history_detail ($hist);
			$bgprouter->lograw (sprintf "Event at %18s -> %s\n",
				strftime ("%D %T", localtime ($ti)), $event);
		}
	}
}

sub process_keepalives_out {
	my $bgprouter = shift;

	my $adjRIBin = $bgprouter->get_adjRIBin ();

	my $sleep_time = 180;
	my $now = time();

	my $key;
	foreach $key ($bgprouter->get_established_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);

		my $keepalive_out = $neighbor->get_timer_keepaliveout ();
		my $holdtime = $neighbor->get_holdtime ();
		if (defined $keepalive_out && defined $holdtime && $holdtime > 0) {
			my $next_keepalive_out = $keepalive_out + ($holdtime / 3) - $now;
			if ($next_keepalive_out <= 3) {
				if (!$neighbor->send_keepalive()) {
					$bgprouter->log (8, 16,
						"Can't send keepalive to $key: $!");
					$neighbor->close ($adjRIBin, -1);
				}
				$next_keepalive_out = $holdtime / 3;
			}
			if ($next_keepalive_out < $sleep_time) {
				$sleep_time = $next_keepalive_out;
			}
		}
	}
	return $sleep_time;
}

sub process_keepalives_in {
	my $bgprouter = shift;

	my $adjRIBin = $bgprouter->get_adjRIBin ();

	my $now = time();
	my $expired_count = 0;
	my $key;
	foreach $key ($bgprouter->get_established_neighbors ()) {
		my $neighbor = $bgprouter->get_neighbor ($key);

		my $keepalive_in = $neighbor->get_timer_keepalivein ();
		my $holdtime = $neighbor->get_holdtime ();

		if (defined $keepalive_in && defined $holdtime && $holdtime > 0) {
			if ($keepalive_in + $holdtime < $now) {
				$expired_count++;
				$bgprouter->log (8, 16, "Holdtimer expired for $key");
				$neighbor->close ($adjRIBin, BGP_ERRC_Hold_Timer_Expired);
			}
		}
	}
	return $expired_count;
}

sub set_loglevel {
	my $bgprouter = shift;
	
	my ($type, $level) = @_;

	$$bgprouter{log}{$type} = $level;
}
	
sub get_loglevel {
	my $bgprouter = shift;

	my $type = shift;

	return $$bgprouter{log}{$type};
}

sub get_defined_loglevels {
	my $bgprouter = shift;

	return (keys %{$$bgprouter{log}});
}

sub set_logging {
    my $bgprouter = shift;
    
    my ($logtype) = @_;
    
    $logtype = lc ($logtype);
    
    setlogsock ('unix');
    if ($logtype eq 'syslog') {
	openlog ('bgpd.pl', 'pid', 'daemon');
		$bgprouter->set_log_filehandle (undef);
    }
    elsif ($logtype eq 'stdout') {
	$bgprouter->set_log_filehandle (\*STDOUT);
    }
# 	elsif ($logtype eq 'file') {
# 		my $fh = new FileHandle;
# 		open ($fh, ">bgpd.log");
# 		$bgprouter->set_log_filehandle ($fh);
# 		$fh->autoflush ();
#	}
else {#else assume file name
	  my $fh = new FileHandle;
	  if($logtype eq "") {$logtype="bgpd.log";} #default log file name
	  open ($fh, ">$logtype");
	  $bgprouter->set_log_filehandle ($fh);
	  $fh->autoflush ();
      }

#   print "set_logging: $logtype.\n";
    $bgprouter->log(128,4,"set_logging: $logtype.\n");
}

sub set_log_filehandle {
	my $bgprouter = shift;

	my $fh = shift;

	if (defined $fh) {
#		$fh->autoflush ();
	}
	return $$bgprouter{loghandle} = $fh;
}


sub get_log_filehandle {
	my $bgprouter = shift;

	return $$bgprouter{loghandle};
}

sub log {
	my $bgprouter = shift;

	my ($type, $priority, @msgs) = @_;

	my $fh = $bgprouter->get_log_filehandle ();

	return if (!defined $fh && $priority > 2);

	my $loglevel = $bgprouter->get_loglevel ($type);

	if ($priority <= 2 || (defined ($loglevel) && $priority <= $loglevel) ) {
		my $message = "";
		if (defined $fh) {
			$message = strftime ("%D %T", localtime (time));
		}
		$message .= sprintf(" %03s/%03s BGP: ", $type, $priority);

		my $level;
		if ($priority == 1 || $priority == 16) {
			$message .= "-ERROR- ";
			$level = 'crit';
		}
		if ($priority == 2 || $priority == 32) {
			$message .= "WARNING ";
			$level = 'warning';
		}
		$message .= join (' ', @msgs);

		if (!defined $fh) {
			syslog ($level, $message);
		} else {
			print $fh "$message\n";
		}
	}
}

sub lograw {
	my $bgprouter = shift;

	my ($message) = @_;
	
	my $fh = $bgprouter->get_log_filehandle ();

	if (!defined $fh) {
		syslog ("info", $message);
	} else {
		print $fh $message;
	}
}

sub beacon {
    my $bgprouter = shift;
    my ($ann) = @_;
    my ($key) = "";
    my ($prefixStr)= $bgprouter->get_beacon_prefix();

    $bgprouter->log(300,4, "ann is equal to $ann, prefix: $prefixStr\n");

    if($ann==1) {
	$bgprouter->log (300, 4, "Sending ANN Beacon: ".localtime()."\n");
	#increment sequence number
	$bgprouter->increment_seqNo(); $bgprouter->increment_ASseqNo();
    }else{
	$bgprouter->log (300, 4, "Sending WD Beacon: ".localtime()."\n");
    }

#    $bgprouter->log (300, 8, "########enter beacon......");

    foreach $key ($bgprouter->get_neighbors()) {
	my $neighbor = $bgprouter->get_neighbor($key);
	my $state = $neighbor->get_state ();
	my $my_ip = $bgprouter->get_routerid ();
	my $my_as = $bgprouter->get_as ();
	my $n_ip=$neighbor->get_ipaddress();

	if($state == BGP_NEIGHBOR_ESTABLISHED) {
#	    if($ann==0) {$bgprouter->log (300, 8, "XXXXX Send a WD.\n");} 
#	    else {$bgprouter->log (300, 8, "XXXXX Send a ANN.\n");} 
	    $neighbor->send_update($bgprouter, $prefixStr, $ann);
	    
	    $bgprouter->log (300, 4, "Beacon seqno(neighbor:$n_ip): ".$bgprouter->get_seqNo().", AGG_AS: ".$bgprouter->get_ASseqNo().($ann==1? " (ANN)":" (WD)"));
	}
    }
#    $bgprouter->log (300, 8, "#########exit beacon......");

}

sub set_beacon_prefix {
	my $bgprouter = shift;
	my $prefix = shift;
	$$bgprouter{beaconPrefix} = $prefix;
}

sub get_beacon_prefix {
	my $bgprouter = shift;
	return $$bgprouter{beaconPrefix};
}
#local pref
sub set_local_pref {
	my $bgprouter = shift;
	my $val = shift;
	$$bgprouter{localPref} = $val;
}

sub get_local_pref {
	my $bgprouter = shift;
	return $$bgprouter{localPref};
}
#MED
sub set_MED {
	my $bgprouter = shift;
	my $val = shift;
	$$bgprouter{MED} = $val;
}

sub get_MED {
	my $bgprouter = shift;
	return $$bgprouter{MED};
}

#community
sub set_community {
	my $bgprouter = shift;
	my $community = shift;
	my @vals = @_;
	my $vals = join('_', @vals);
	$$bgprouter{community} = $vals;
	$bgprouter->log(300,8,"community values: @vals: $#vals, $vals");
}

sub get_community {
	my $bgprouter = shift;
	return $$bgprouter{community};
}
#confederation ID
sub set_confedID {
	my $bgprouter = shift;
	my $val = shift;
	$$bgprouter{confedID} = $val;
$bgprouter->log(300,8,"AS_CONFED: setting confed ID: $val");
}

sub get_confedID {
	my $bgprouter = shift;
	return $$bgprouter{confedID};
}

#confederation peers: bgp confederation peers 65000 xxxx xxxx ...
sub set_confedPeers {
	my $bgprouter = shift;
	shift; shift; shift;
	my @vals = @_;
	my $val = join('_', @vals);
	$$bgprouter{confedPeers} = $val;

$bgprouter->log(300,8,"AS_CONFED: setting confed peers: $val");
}

sub get_confedPeers {
	my $bgprouter = shift;
	return $$bgprouter{confedPeers};
}

sub isSameConfed {
    my $bgprouter = shift;
    my $val = shift;
    my $confedPeers = $bgprouter->get_confedPeers();
    #is it one of confed peers?
    if($confedPeers eq $val || $confedPeers=~/^${val}_/ || 
       $confedPeers=~/_${val}_/ || $confedPeers=~/_$val$/) {
	return 1;
    }
    return 0;
}


#need to keep per Neighbor sequence number:
#sequence number from 0 to 65535
sub increment_seqNo {
    my $bgprouter = shift;
    my $seqNo = $$bgprouter{seqNo};
    $$bgprouter{seqNo} = ($seqNo==65535)?  1 : ($seqNo+1);
}

sub set_seqNo {
    my $bgprouter = shift;
    my $seqNo = shift;
    $$bgprouter{seqNo} = $seqNo;
}

sub get_seqNo {
    my $bgprouter = shift;
    return $$bgprouter{seqNo};
}

#AS sequence number from 64512 to 65535
sub increment_ASseqNo {
    my $bgprouter = shift;
    my $ASseqNo = $$bgprouter{ASseqNo};
    $$bgprouter{ASseqNo} = ($ASseqNo==65535)?  64512 : ($ASseqNo+1);
}

sub set_ASseqNo {
    my $bgprouter = shift;
    my $ASseqNo = shift;
    $$bgprouter{ASseqNo} = $ASseqNo;
}

sub get_ASseqNo {
    my $bgprouter = shift;
    return $$bgprouter{ASseqNo};
}
